(in-package :asn.1)

(deftype access () '(member :not-accessible
                            :accessible-for-notify
                            :read-only
                            :read-write
                            :read-create))

(deftype status () '(member :current :deprecated :obsolete))

(deftype type-specifier () '(or symbol list))

(defclass object-id (asn.1-type)
  ((name        :type string
                :reader oid-name
                :initarg :name)
   (value       :type integer
                :reader oid-value
                :initarg :value)
   (type        :type symbol
                :reader oid-type
                :initarg :type)
   (syntax      :type type-specifier
                :reader oid-syntax
                :initarg :syntax)
   (max-access  :type access
                :reader oid-max-access
                :initarg :max-access)
   (status      :type status
                :reader oid-status
                :initarg :status)
   (description :type string
                :reader oid-description
                :initarg :description)
   (module      :type symbol
                :reader oid-module
                :initarg :module)
   (parent      :type object-id
                :reader oid-parent
                :initarg :parent)
   (children    :type hash-table
                :accessor oid-children
                :initform (make-hash-table)))
  (:documentation "OBJECT IDENTIFIER"))

(defmethod print-object ((object object-id) stream)
  (print-unreadable-object (object stream :type t)
    (cond ((and (slot-boundp object 'name)
                (slot-boundp object 'value))
           (format stream "~A (~D) [~D]"
                   (oid-name object)
                   (oid-value object)
                   (hash-table-count (oid-children object))))
          ((slot-boundp object 'value)
           (format stream "~A.~D"
                   (oid-name (oid-parent object))
                   (oid-value object))))))

(defmethod initialize-instance :after ((instance object-id)
                                       &rest initargs &key &allow-other-keys)
  (declare (ignore initargs))
  (if (and (slot-boundp instance 'parent)
           (slot-boundp instance 'value))
      (let ((o (oid-parent instance)))
        (setf (gethash (oid-value instance) (oid-children o)) instance))))

(defun list-children (oid)
  (declare (type object-id oid))
  (let ((all-children '()))
    (maphash #'(lambda (key value)
                 (declare (ignore key))
                 (push value all-children))
             (oid-children oid))
    all-children))

(defun delete-object (oid)
  (declare (type object-id oid))
  (setf (gethash (oid-value oid)
                 (oid-children (oid-parent oid)))
        nil))

(defun oid-parent-p (oid)
  (declare (type object-id oid))
  (slot-boundp oid 'parent))

(defun oid-number-list (oid)
  (declare (type object-id oid))
  (labels ((iter (o acc)
             (if (slot-boundp o 'parent)
                 (iter (oid-parent o) (cons (oid-value o) acc))
               acc)))
    (iter oid nil)))

(defun oid-name-list (oid)
  (declare (type object-id oid))
  (labels ((iter (o acc)
             (if (slot-boundp o 'parent)
                 (iter (oid-parent o) (cons (oid-name o) acc))
               acc)))
    (iter oid nil)))

(defun oid-length (oid)
  (declare (type object-id oid))
  (list-length (oid-number-list oid)))

(defmethod plain-value ((object object-id))
  (coerce (oid-number-list object) 'vector))

(defmethod ber-equal ((a object-id) (b object-id))
  (or (equal (oid-number-list a) (oid-number-list b))
      (and (equal (oid-number-list a) '(0)) (equal (oid-number-list b) '(0 0)))
      (and (equal (oid-number-list b) '(0)) (equal (oid-number-list a) '(0 0)))))

;;; BER Encode & Decode (:object-identifier)
(defmethod ber-encode ((value object-id))
  (labels ((number-get (n)
             (if (zerop n) (values (list 0) 1)
               (number-split n 0 nil 0)))
           (number-split (n p acc l)
             (if (zerop n) (values acc l)
               (multiple-value-bind (q r) (floor n 128)
                 (number-split q 1 (cons (logior (ash p 7) r) acc) (1+ l)))))
           (iter (oids acc len)
             (if (endp oids)
                 (values acc len)
               (multiple-value-bind (sub-oid sub-length) (number-get (car oids))
                 (iter (cdr oids) (nconc acc sub-oid) (+ len sub-length))))))
    (let ((subids (oid-number-list value)))
      (multiple-value-bind (v l)
          (case (oid-length value)
            (0 (values nil 0))
            (1 (number-get (* (first subids) 40)))
            (2 (number-get (+ (* (first subids) 40) (second subids))))
            (otherwise (apply #'iter
                              (cddr subids)
                              (multiple-value-list
                               (number-get (+ (* (first subids) 40)
                                              (second subids)))))))
        (concatenate 'vector
                     (ber-encode-type 0 0 6)
                     (ber-encode-length l)
                     v)))))

;;; FIXME: #(6 1 40), |iso| should be decoded to 1, not 1.0
(defmethod ber-decode-value ((s stream) (type (eql :object-identifier)) length)
  (declare (type stream s)
           (type fixnum length)
           (ignore type))
  (if (zerop length)
      (oid '(0))
    (labels ((get-number (acc len)
               (let* ((byte (read-byte s))
                      (val (logior (ash acc 7) (logand byte 127))))
                 (if (< byte 128) (values val len)
                   (get-number val (1+ len)))))
             (iter (left-length acc head-p)
               (declare (type fixnum left-length)
                        (type list acc))
               (if (zerop left-length)
                   (nreverse acc)
                 (multiple-value-bind (n l) (get-number 0 1)
                   (if head-p
                       (multiple-value-bind (q r) (floor n 40)
                         (iter (- left-length l) (cons r (cons q acc)) nil))
                     (iter (- left-length l) (cons n acc) nil))))))
      (oid (iter length nil t)))))

(eval-when (:load-toplevel :execute)
  (install-asn.1-type :object-identifier 0 0 6))

;;; Object ID utilities, from Lisp-SNMP
(defun oid-< (oid-1 oid-2)
  "test if oid-1 is oid-2's child"
  (let ((o-1 (nreverse (oid-number-list oid-1)))
        (o-2 (nreverse (oid-number-list oid-2)))
        (o-1-len (oid-length oid-1))
        (o-2-len (oid-length oid-2)))
    (if (<= o-1-len o-2-len) nil
      (equal o-2 (nthcdr (- o-1-len o-2-len) o-1)))))

(defun oid->= (oid-1 oid-2)
  (not (oid-< oid-1 oid-2)))

(defvar *root-object-id*
  (make-instance 'object-id))

(defvar |zero|
  (make-instance 'object-id
                 :name "zero"
                 :value 0
                 :parent *root-object-id*))

(defvar |iso|
  (make-instance 'object-id
                 :name "iso"
                 :value 1
                 :parent *root-object-id*))

(defmacro defoid (name (parent value))
  `(defvar ,name (make-instance 'object-id
                                :name (symbol-name ',name)
                                :value ,value
                                :parent ,parent)))

;;; Resolver Macro

;; sysDescr.0 -> |sysDescr| and 0
;; 1.3.6.1.2.1.1.1.0 -> *root-object-id*, 1, 3, 6, 1, 2, 1, 1, 1, 0
(defvar *oid-readtable*)

(defun make-oid-readtable ()
  ;; Create a new readtable
  (setf *oid-readtable* (copy-readtable nil))
  ;; Case-Sensitivity
  (setf (readtable-case *oid-readtable*) :preserve)
  ;; . reader
  (set-macro-character #\. #'(lambda (stream char)
                               (declare (ignore stream char))
                               (values))
                       nil *oid-readtable*)
  (set-macro-character #\} (get-macro-character #\))
                       t *oid-readtable*))

(eval-when (:load-toplevel :execute)
  (make-oid-readtable))

(defgeneric oid-parse (source))

(defmethod oid-parse ((source string))
  (with-input-from-string (s source)
    (oid-parse s)))

(defmethod oid-parse ((source stream))
  (let ((*readtable* *oid-readtable*)
        (*package* (find-package :asn.1)))
    (loop for i = (read source nil nil nil)
          until (null i)
          collect i)))

(defgeneric oid (object)
  (:documentation "Convert * to OBJECT-ID"))

(defmethod oid ((oid-object object-id))
  oid-object)

(defmethod oid ((oid-vector vector))
  (oid (coerce oid-vector 'list)))

(defmethod oid ((oid-tokens list))
  (labels ((iter (tokens base)
             (if (null tokens)
                 base
               (let ((i (car tokens))
                     (b (gethash (car tokens)
                                 (oid-children base))))
                 (etypecase i
                   (integer (iter (cdr tokens)
                                  (if b b
                                    (make-instance 'object-id
                                                   :value i
                                                   :parent base))))
                   (symbol (iter (cdr tokens)
                                 (symbol-value i))))))))
    (etypecase (car oid-tokens)
      (integer (iter oid-tokens *root-object-id*))
      (symbol (iter (cdr oid-tokens)
                    (symbol-value (car oid-tokens)))))))

(defmethod oid ((oid-string string))
  (oid (oid-parse oid-string)))

(defvar *possible-oid-reader-function* nil)

(defun oid-reader (stream char1 char2)
  (declare (ignore char1 char2))
  (let ((tokens-list
         (loop for i = (read-char stream nil #\} t)
               until (char= i #\})
               collect i)))
    (oid (coerce tokens-list 'string))))

(defun enable-asn.1-user-readtable ()
  (push (get-dispatch-macro-character #\# #\{)
        *possible-oid-reader-function*)
  (set-dispatch-macro-character #\# #\{ #'oid-reader))

(defun disable-asn.1-user-readtable ()
  (set-dispatch-macro-character #\# #\{
                                (pop *possible-oid-reader-function*)))
